import networkx as nx
import itertools as it
class ReactionParameters:
  def __init__(self,substrate,product,km_fwd,km_back,kcat_fwd,kcat_back):
    self.substrate = substrate
    self.product = product
    self.km_fwd = km_fwd
    self.km_back = km_back
    self.kcat_fwd = kcat_fwd
    self.kcat_back = kcat_back

class Enzyme:

  def __init__(self,name,amount):
    self.name = name
    self.amount = amount
    self.reactions = dict()
  
  def add_reaction_params(self,r):
    self.reactions[r.substrate.name+"."+r.product.name] = r
  
class Reactant:
  def __init__(self,name,amount):
    self.name = name
    self.amount = amount

  def __str__(self):
    return self.name

class Experiment:

  def __init__(self):
    self.enzymes = dict()
    self.reactants = dict()


  def add_enzyme(self,enzyme):
    self.enzymes[enzyme.name] = enzyme
  
  def add_reactant(self,reactant):
    self.reactants[reactant.name] = reactant
  
  def _build_reaction_graph(self):
    reaction_graph = nx.DiGraph()
    for k in self.reactants.keys():
      reaction_graph.add_node(self.reactants[k])
    for k,e in self.enzymes.items():
      for name,rt in e.reactions.items():
        reaction_graph.add_edge(rt.substrate,rt.product,enzyme=e)
    return reaction_graph

  # def reversible_mm(self,Vf_max,S,Km_forward,Vb_max,P,Km_back):
  def reversible_mm(self,enzyme,substrate,product,params):
    Km_forward = params.km_fwd
    Km_back = params.km_back
    Vf_max = params.kcat_fwd * enzyme.amount
    Vb_max = params.kcat_back * enzyme.amount
    S = substrate.amount
    P = product.amount

    nom = (Vf_max * S / Km_forward ) - Vb_max * P / Km_back
    denom = 1 + S/Km_forward + P/Km_forward
    return nom/denom

  def iteration(self,reaction_graph,source,dt):
      for rsource, rtarget in it.product(reaction_graph.nodes,reaction_graph.nodes):
        if (rsource.amount == 0 and rtarget.amount == 0) or (rsource == rtarget):
          continue
        if reaction_graph.has_edge(rsource,rtarget):
          enzyme = reaction_graph[rsource][rtarget]['enzyme']
          r_params = enzyme.reactions[rsource.name+'.'+rtarget.name]
          v = self.reversible_mm(enzyme,rsource,rtarget,r_params)
          
          #update the amounts:
          rsource.amount -= max(v*dt,0)
          rtarget.amount += max(v*dt,0)

  def run(self,steps,dt=0.1):
    reaction_graph = self._build_reaction_graph()
    reactants = sorted(list(self.reactants.values()),key=lambda x: -x.amount)
    source = reactants[0]
    results = []
    for i in range(steps):
      self.iteration(reaction_graph,source,dt)
      record = {'step':i}
      for name,obj in self.reactants.items():
        record[name] = obj.amount
      results.append(record)
    return results



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'src'}

packages = \
['kinetix']

package_data = \
{'': ['*']}

install_requires = \
['matplotlib>=3.1.3',
 'networkx>=2.5',
 'pandas>=1.1.5',
 'seaborn>=0.12.0',
 'yaml>=6.0']

entry_points = \
{'console_scripts': ['kinetix = kinetix.runner:main']}

setup_kwargs = {
    'name': 'kinetix',
    'version': '0.1.0b0',
    'description': 'A simple lightweight library for simulating and plotting enzyme kinetics',
    'long_description': '# Kinetix - A simple enzyme kinetics simulation framework\n\nThis is a simple framework for single / multiple reaction enzyme kinetics simulation and plotting.\nIt allows one to plot reaction kinetics that follows the reversible [Michaelis - Menten](https://en.wikipedia.org/wiki/Michaelis%E2%80%93Menten_kinetics) model.\n\n## Usage\nOne can use the library to produce figures via a command line interface or programmatically (see below for examples)\n\n### CLI\n#### Single reaction\n\nAs an example, the reaction kinetics of Glucose Kinase (EC: [2.7.1.2](https://www.brenda-enzymes.org/enzyme.php?ecno=2.7.1.2) is displayed as an example)\n<p align="center">\n  <img src="https://upload.wikimedia.org/wikipedia/commons/8/8d/Glucokinase.png" />\n</p>\nFirst, create a definition file in YAML format, that defines the reaction parameters:\n\n```yaml\nname: "Glucose kinase simulation"\nreactants: #Concentration of the reactants at the beginning of the reaction (mM)\n  glucose: 0.2\n  glucose_6_p: 0.0\nenzymes: #Concentration of the enzyme(s) at the begining of the reaction (mM)\n  gluk: 0.05\nreaction: #Kinetic parameters of each of the enzymes for the forward and backward steps of the (reversible) reaction. Note that non-reversible reactions can simple be modeled with high Km for on of the directions.\n  gluk:\n    fwd: "glucose_6_p"\n    back: "glucose"\n    km_fwd: 0.24\n    km_back: 21\n    kcat_fwd: 61\n    kcat_back: 15.9\n```\n\nThen, run the simulation with a simple command line and a few arguments:\n```\npython ./code/runner.py examples/glucose_kinase.yaml --plot_out gluc.png --csv_out gluc.csv\n```\nA figure showing the progression of the reaction as a function of time is generated:\n<p align="center">\n  <img src="examples/figures/gluc.svg" />\n</p>\n\nA csv file containing the data used to generate the figure can also optionally be generated and saved (using the `--csv_out` flag)\n\n#### A pathway\nKinetix can also simulate a pathway composed of several enzymes. This example shows a pathway composed of 3 different enzymes used in the production of [allulose](https://en.wikipedia.org/wiki/Psicose) (D-psicose) a C3 epimer of fructose:\n\n1. **Fructose kinase** (for the production of fructose-6-phosphate)\n2. **D-psicose-3-epimerase** (converting fructose-6-phosphate to allulose-6-phosphate)\n3. **Alkaline phosphatase** (converting allulose-6-phosphate to allulose)\n\n\nThe flow is similar to the one-enzyme case. First, define a yaml file with all the parameters:\n\n```yaml\n---\nname: "Allulose synthesis from fructose"\nreactants:\n  fructose: 1.8\n  fructose_6_p: 0.0\n  allulose: 0.0\n  allulose_6_p: 0.0\nenzymes:\n  fruk: 0.05\n  alsE: 0.05\n  phosphatase: 0.05\nreaction:\n  fruk:\n    fwd: "fructose_6_p"\n    back: "fructose"\n    km_fwd: 0.24\n    km_back: 21\n    kcat_fwd: 61\n    kcat_back: 15.9\n  alsE:\n    fwd: "allulose_6_p"\n    back: "fructose_6_p"\n    km_fwd: 1.6 \n    km_back: 1.6 \n    kcat_fwd: 46\n    kcat_back: 46\n  phosphatase:\n    fwd: "allulose"\n    back: "allulose_6_p"\n    km_fwd: 1 \n    km_back: 100\n    kcat_fwd: 100\n    kcat_back: 1\n```\n\nThen, invoke the application just as before:\n```\n./code/runner.py examples/allulose.yaml --plot_out examples/figures/alu.svg --csv_out examples/csvs/alu.csv\n```\n\nTo produce the reaction figure, which displays the concentration of each of the reactants as a function of time:\n<p align="center">\n  <img src="examples/figures/alu.svg" />\n</p>\n\n### Programmatic Access\nKinetix also makes it really easy to simulate and plot a reaction more flexibly via a simple API.\nExample:\n\nhttps://github.com/LiorZ/Kinetix/blob/4d1107d3e85c45d8d3f222428936ef82c658f223/examples/programmatic.py#L8-L44\n',
    'author': 'Lior Zimmerman',
    'author_email': 'zimmerman.lior@gmail.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'None',
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
}


setup(**setup_kwargs)



import yaml
from .base import *
class KinetixParser:
    def __init__(self, filename):
        self.filename = filename
        self.data = None
        self.experiment = None

    def parse_experiment(self):
        with open(self.filename, 'r') as f:
            self.data = yaml.safe_load(f)
        self.reaction_name = self.data['name']
        self.reactants = dict()
        for rname,rvalue in self.data['reactants'].items():
            self.reactants[rname] = Reactant(rname, rvalue)
        self.enzymes = dict()
        for ename,econc in self.data['enzymes'].items():
            self.enzymes[ename] = Enzyme(ename, econc)
            if ename not in self.data['reaction'].keys():
                break
            params = self.data['reaction'][ename]
            reaction_param = ReactionParameters(self.reactants[params['back']],self.reactants[params['fwd']],params['km_fwd'],params['km_back'],params['kcat_fwd'],params['kcat_back'])
            self.enzymes[ename].add_reaction_params(reaction_param)
        self.experiment = Experiment()
        for e in self.enzymes.values():
            self.experiment.add_enzyme(e)
        for r in self.reactants.values():
            self.experiment.add_reactant(r)
        return self.experiment

    def get_experiment():
        return self.experiment








#!/usr/bin/env python

from kinetix import parser
import argparse 
import sys 
import pandas as pd
import matplotlib.pyplot as plt
import seaborn as sns

def run(exp,args):
    result = exp.run(args.steps,dt=args.delta_t)

    df = pd.DataFrame(result)
    if args.plot_out is not None:
        ax = sns.lineplot(data=df.melt(id_vars='step'),y='value',x='step',hue='variable')
        ax.set_xlabel('Time (100 ms increments)')
        ax.set_ylabel('Concentration (mM)')
        plt.savefig(args.plot_out)
    if args.csv_out is not None:
        df.to_csv(args.csv_out,index=False)


def main():
    argparser = argparse.ArgumentParser()
    argparser.add_argument('input', help='input yaml file with reaction definitions')
    argparser.add_argument('--csv_out', help='output csv file with reaction steps',default=None)
    argparser.add_argument('--plot_out', help='plot of the reaction kinetics',default=None)
    argparser.add_argument('--steps', help='number of steps to simulate (seconds)', type=int, default=1000)
    argparser.add_argument('--delta_t',help='size of each step (seconds)', type=float, default=0.01)
    args = argparser.parse_args()

    # Read the yaml file:
    yaml_parser = parser.KinetixParser(args.input)
    
    experiment = yaml_parser.parse_experiment()
    run(experiment,args)
    
if __name__ == '__main__':
    main()
